--****
-- == Windows Task Dialogs
--
-- A task dialog is similar to a Unicode message box with several extra features.\\
-- Like a message box, it is formatted by the operating system.
-- [[:abouttaskdialogboxes|More]]
--
-- <<LEVELTOC level=2 depth=4>>
--

namespace TaskDlg		-- Euphoria version 4 or later is required.

-- By Arthur Robert Crump, 2013, email arthur.crump@
-- ntlworld.com   Documentation is in Eudoc/Creole format within this file.

atom TDF_default	-- Default Flags for TaskDialog. The value of TDF_default will be set
					-- immediately after the Task dialog flags (TDF_*) have been declared.
					-- The default is in the documentation, so don't change it on its own.

include std/error.e
ifdef not WINDOWS then
	crash("Only available in Windows systems (Vista or later)")
end ifdef

constant WM_USER = #0400,
		comctl32 = open_dll("comctl32"),
		user32 = open_dll("user32.dll"),
		TaskDlgInd = define_c_func(comctl32,"TaskDialogIndirect",
					{C_POINTER, C_POINTER, C_POINTER, C_POINTER}, C_ULONG),
		SendMsg = define_c_func(user32,"SendMessageW",
					{C_HWND, C_UINT, C_UINT, C_LONG}, C_LONG)

if TaskDlgInd < 0 then
	crash("Task dialogs are only available on Windows Vista or later")
end if

include std/dll.e
include std/types.e
include std/os.e
include std/machine.e
include std/math.e

--		Defining the fields of the TASKDIALOGCONFIG structure
enum by +4 cbSize=0, hwndParent, hInstance, dwFlags, dwCommonButtons,
			pszWindowTitle, hMainIcon, pszMainInstruction, pszContent,
			cButtons,pButtons, nDefaultButton, cRadioButtons, pRadioButtons,
			nDefaultRadioButton, pszVerificationText, pszExpandedInformation,
			pszExpandedControlText, pszCollapsedControlText, hFooterIcon,
			pszFooter, pfCallback, lpCallbackData, cxWidth, SIZE_TASKDIALOGCONFIG
constant pszMainIcon=hMainIcon, pszFooterIcon=hFooterIcon

atom NextPtr			-- Where to store next pointer in a TASKDIALOGCONFIG extension

--****
-- === About task dialog boxes
--
-- A task dialog can have more general icons, messages and buttons than a message box.
-- In addition, a task dialog may have a set of radio buttons, command links,
-- a progress bar and one checkbox, described as a verification checkbox.
-- Also, a callback routine may be specified to receive notification messages.\\
-- A task dialog is usually created using the Euphoria function [[:TaskDialog]]. 
-- 
-- Task dialogs are independent of any other GUI (Graphic User Interface) software and
-- may be used in a program without any other GUI or which also uses other GUI software
-- such as win32lib or wxEuphoria without interfering with its operation except to pause
-- it while the task dialog is open. For example, a task dialog could be activated before
-- starting another GUI and, later, when the other GUI is running, another task dialog
-- could be activated from an event handler.
--
-- All text should be in 16-bit Unicode (UTF-16). The routine does not translate from
-- 8-bit codes, but Euphoria sequences containing 7-bit ASCII should not need translation.
-- A few extended ASCII 8-bit codes may need translation, particularly the **euro**
-- currency symbol. Unicode characters in the supplementary planes (UTF-32 above #FFFF) 
-- will be detected and displayed as unknown characters since they are not supported by 
-- the typeface used by task dialogs.
--
-- More information on task dialog boxes may be found in the Microsoft documentation for the
-- [[http://msdn.microsoft.com/en-us/library/windows/desktop/bb760544(v=vs.85).aspx "TaskDialogIndirect API"]]
-- and the
-- [[TASKDIALOGCONFIG -> http://msdn.microsoft.com/en-us/library/windows/desktop/bb787473(v=vs.85).aspx ]]
-- structure. See also the simpler
-- [[TaskDialog API -> http://msdn.microsoft.com/en-us/library/windows/desktop/bb760540(v=vs.85).aspx]]
--

--****
-- ==== Limitations
--
-- * Task dialogs are only available on Windows Vista or later.
-- * Task dialogs cannot contain any controls not mentioned above, such as combo boxes,
-- list boxes or edit boxes for the input of text or numeric data.
-- * Only one group of radio buttons is supported.
-- * There are no options for changing the typeface or its size.

-- The example below uses spaces instead of tabs.

--****
-- ==== Example
--
-- A task dialog can be created quite easily. 
-- The following example is the first example in the Examples folder.
--
-- <eucode>
-- include TaskDialog.e
--
-- object Result = TaskDialog(0,                -- No parent window 
--      ,                                       -- Flags omitted - using default
--      "Simple test",                          -- Title
--      "Click any button or radio button",     -- The main instruction
--              -- The next two lines are the Content giving more information
--      "This is a simple test dialog to check the TaskDialog wrapper."&
--      "\nThe euro symbol "&#20AC&" is hexadecimal 20AC",
--              -- the next line is the footer text.
--      #300A&" Footer text goes here "&#300B,
--              -- Next the main icon (none) and footer icon
--      {0,{TD_INFORMATION_ICON}},
--              -- The next two lines are the buttons and the radio buttons
--      {TDCBF_OK_BUTTON,TDCBF_CANCEL_BUTTON,{100,"Maybe"},{200,#BF&"Che?"},{IDOK}},
--      {{101,"English"},{102,"Swedish"},{103,"Vanish"},{104,"Klingon"},{102}},
--              -- The next two lines are expanded text
--      {"This is 'expanded text' which is usually not shown\nuntil"&
--      " it is expanded by pressing 'Show More information'",
--              -- then the two control texts continuing the same parameter
--      "Show more information","Collapse the extra information"},
--      "Do you want to click this checkbox?")       -- Verification checkbox text
--              -- There is no callback procedure, so the parameter is omitted.
-- </eucode>
--
-- which produces:
--
-- {{images/ExampleDoc.png|Cannot display image - see image/ExampleDoc.png}}
--

-- Local routines - Not in eudoc/creole documentation.

-- Parameters of the UTF16 function when called by TaskDialog, TaskDialogIndirect
-- or SendMessageW:
--
-- 'Where' : Specifies a location to store the result of processing the 'str' paramaeter.
-- If 'Where' is 0 the result of processing the 'str' parameter is returned as the
-- main result instead of being stored.
--
-- 'str' may be an atom or a sequence. If 'str' is an atom it may be a pointer to an
-- existing null-terminated unicode string or an integer resource identifier or 0.
-- This would be stored in 'Where' without further checks and a success code 16
-- returned unless 'Where' is 0, in which case it is returned as the main result.
--
-- If 'str' is a sequence there is a basic check that its elements are at least in the
-- range of unicode characters. The string is then stored as a null-terminated 16-bit
-- unicode string. Its pointer is stored in 'Where' or returned as the main result if
-- 'Where' is 0.
--
-- If 'Where' is non-zero the function returns one of the following:
-- 0	to indicate that the sequence 'str' is not a valid unicode string, or
-- 16	to indicate that 'str' is a valid UTF-16 string or an atom
-- 32	to indicate that 'str' contained UTF-32 character(s) which have been converted
--		into UTF-16 surrogate pairs. This conversion has not been well tested and is
--		not in the Creole documentation. Furthermore the API uses a typeface 'Segoe UI'
--		which does not support them and they would be shown as an empty rectangle.
--		However, they will not cause TaskDialog to fail.
-- If 'Where' is 0, the function returns the result of processing 'str', 
-- which may be 0 if 'str' is an invalid sequence.

function UTF16( atom Where, object str )
	integer temp, rtn = 16
	atom ptr
	if atom(str) then
		ptr = str
	else
		for n = length(str) to 1 by -1 do
			if sequence(str[n]) or str[n] > #10FFFF then
				return 0			-- Not a Unicode character
			elsif str[n] > #FFFF then
				temp = str[n] - #10000
				str = insert(str,and_bits(temp,#3FF)+#DC00,n+1)
				str[n] = floor(temp/1024)+#D800
			end if
		end for
		-- If it gets here it is OK.
		ptr = allocate_wstring(str)
		if Where then
			-- Store the result and return 16 for UTF16 or 32 if any UTF32 found 
			poke4( NextPtr, ptr )		-- Remember it for deallocation later.
			NextPtr += 4
			poke4( Where, ptr )
			-- If Where is 0, only the check is performed.
		else
			return ptr
		end if
	end if
	return rtn
end function



--****
-- === Task dialog creation
--

--**
-- The normal behaviour of this routine is to create a configuration structure defining 
-- a task dialog and then display the task dialog. The configuration structure is 
-- defined in the Microsoft documentation 
-- [[TASKDIALOGCONFIG -> http://msdn.microsoft.com/en-us/library/windows/desktop/bb787473(v=vs.85).aspx ]].
-- The parameters of TaskDialog correspond to the fields of that structure but are not 
-- in the same order and some parameters correspond to more than one field,
-- \\For an alternative use of this function see the [[:Multi-page dialogs]] section.
--
-- Parameters:
--
-- # ##hWndInst## : Usually an atom ##hParent##. However, a sequence of two atoms
-- {##hParent##,##hInstance##} is permitted but not recommended.\\
-- ** ##hParent## is a handle to the parent window or 0 if not called from a window.
-- ** ##hInstance## is not usually necessary since the routine can usually obtain the 
-- appropriate handle (see note 1 below).
-- # ##Flags## : an object. A combination of the flags listed in [[:Task dialog flags]]
-- or a sequence of those flags, which will be combined. If ##Flags## is omitted the
-- default is TDF_SIZE_TO_CONTENT.
-- # ##Title## : The task dialog title. Usually a Unicode string (but see note 2).
-- If ##Title## is 0 the filename of the executable program is used. This will be the
-- name of the Euphoria interpreter unless the program is compiled or bound.
-- # ##MainText## is the main instruction and is usually a Unicode string (see notes 2).
-- # ##Content_W## : ##Content## or {##Content##,{##Width##}}, but not {##Content##}.
-- If ##Width##  is specified it must be in a subsequence to avoid confusion with a
-- rather unlikely ##Content_W## as a two-character string.
-- ** ##Content## is usually a Unicode string (see notes 2).
-- ** ##Width## is the width of the task dialog's client area in dialog units.
--  Usually omitted or 0 leaving the task dialog manager to set the width (see note 4)
-- # ##Footer## is usually 0 or a Unicode string to appear in the footer area of 
-- the task dialog (see notes 2).
-- # ##Icons## : ##MainIcon## or {##MainIcon##,##FootIcon##} but not {##MainIcon##}.
-- ** ##MainIcon## is either a handle of an icon or a sequence containing only one element 
-- which is usually one of the [[:Predefined icons]] (see note 3).
-- If ##MainIcon## is the handle of an icon the flag [[:TDF_USE_HICON_MAIN]] must be set.
-- ** ##FootIcon## is either a handle of an icon or a sequence containing only one element
-- which is usually one of the [[:Predefined icons]] (see note 3).
-- If ##FootIcon## is the handle of an icon the flag [[:TDF_USE_HICON_FOOTER]] must be set.
-- # ##Buttons## : A sequence. Each element of the sequence is one of:
-- ** An atom consisting of a combination of [[:Common button flags]]. If there is more
-- than one atom they will be combined.
-- ** A sequence of two elements {button-ID,button-name} where button-name is usually a
-- Unicode string (see note 2). Custom buttons will appear before the common buttons.
-- If no common or custom buttons are specified an "OK" button will appear.
-- ** A sequence consisting of a single atom specifying the ID of the default button.
-- If absent or not a valid ID, the first button shown in the dialog will be chosen
-- (see note 5).
-- # ##Radios## : A sequence, which may be empty. Each element of the sequence is one of:
-- ** A sequence of two elements {button-ID,button-name} where button-name is usually a
-- Unicode string (see note 2).
-- ** A sequence containing a single atom specifying the ID of the radio button to be
-- selected by default. If absent or not a valid ID, the first radio button will be
-- selected unless [[:TDF_NO_DEFAULT_RADIO_BUTTON]] is set (see note 5).
-- # ##ExpandCC## : A sequence of upto three elements, each of which is usually a Unicode
-- string (see note 2).
-- ** The first element defines text to be shown below the content or the footer area,
-- depending upon the flag [[:TDF_EXPAND_FOOTER_AREA]]. If no expanded text has been 
-- specified any other elements of ##ExpandCC## are ignored.
-- ** The second element labels the button to be used to collapse the expanded text.
-- ** The third button labels the button to show the expanded text.\\
-- If only one of the labels is specified it is used for both labels. If neither is 
-- specified, default labels are used.
-- # ##Verify## : Usually a Unicode string labelling a checkbox for verification 
-- (see note 2). If ##Verify## is 0 the checkbox is not displayed.
-- # ##CallBack## : An atom ##RID## or a sequence of upto three elements 
-- {##RID##,##CBdata##,##DontDisplay##} which are explained in [[:Callback support]].
-- If ##RID## is 0 the ##CallBack## parameter is ignored and there is no 
-- callback routine. 
-- 
-- Notes:
--
-- The following features of the Windows API are unlikely to be required, but are
-- supported by this routine
-- 
-- # If ##hInstance## is specified it should be a handle to the module containing any
-- resources used to supply icons and Unicode text strings. If a 0 value of hinstance
-- is specified it will prevent the routine from finding the appropriate handle and any
-- icons and text strings which depend upon it will not be used.
-- # Whenever a parameter or sub-parameter is "Usually a Unicode string" an alternative
-- is a pointer to a previously allocated area containing a null terminated Unicode string
-- (or it could be an integer resource identifier, which should have been obsolete
-- long ago, so avoid these if possible).
-- # ##MainIcon## and ##FootIcon## may also be a sequence containing an Integer Resource
-- Identifier instead of one of the Predefined icons. To be avoided if possible.
-- # If a client width is not specified in ##Content_W## and [[:TDF_SIZE_TO_CONTENT]]
-- is not set, a default width is set, which may truncate long strings without spaces,
-- such as file names or hyperlinks, to approximately 48 characters unless there are
-- enough buttons or other controls to increase the width.
-- # If more than one default button or radio button is specified, the later selection
-- will override any earlier selections. The default radio button may also be specified
-- as an atom, since there are no common radio buttons.
--
-- Returns:
-- 
-- Assuming that ##DontDisplay## is omitted, or set to its default value of 0,
-- TaskDialog returns a sequence of four elements.
--
-- If an argument error was detected before calling [[:TaskDialogIndirect]]~:
--
-- # ##ReturnCode## : E_INVALIDARG (see [[:Return Codes]] )
-- # ##Reason## : A string with more information
-- # ##Fill1## : 0
-- # ##Fill2## : 0			
--
-- Otherwise, TaskDialog returns results from [[:TaskDialogIndirect]]~:
-- # ##ReturnCode## : One of the following [[:Return Codes]]
-- ** S_OK
-- ** E_OUTOFMEMORY
-- ** E_INVALIDARG
-- ** E_FAIL
-- # ##ButtonID## : The button ID of a selected button, which may be one of the custom
-- buttons or one of the following common buttons: see [[:Return codes]]
-- ** IDCANCEL
-- ** IDNO
-- ** IDOK
-- ** IDRETRY
-- ** IDYES
-- # ##Radio## : The button ID of a selected radio button, or 0 if there are no radio
-- buttons or none have been selected.
-- # ##Verified## : True if the verification checkbox is set, False if it is clear or
-- there is no checkbox
--
--
-- If ##DontShow## has been set, the return value is a pointer to the configuration 
-- structure so that the task dialog can be created later; see [[:Multi-page dialog]]
--


public function TaskDialog( object hWndInst=0, object Flags=TDF_default, object Title=0, 
			object MainText, object Content_W=0, object Footer=0, object Icons=0, 
			sequence Buttons={}, sequence Radios={}, sequence ExpandCC={}, 
			object Verify=0, object CallBack=0 )
	atom config,		-- Pointer to memory block to hold TASKDIALOGCONFIG structure 
		 flg,			-- To hold various flags temporarily 
		 Bptr			-- Pointer to memory to hold custom buttons, then radio buttons
	sequence Result		-- See Microsoft TaskDialogIndirect documentation
	integer cBtn,		-- Count of buttons, then count of radio buttons
			ConfigSize	-- The size of the config area extended by an area for pointers.
	ConfigSize = SIZE_TASKDIALOGCONFIG + 4*length(Buttons&Radios) + 48
	-- Allocate memory for the structure with extra space for pointers to other memory
	-- areas which may be cleared by FreeConfig later. This may be more than required
	-- but if it has been calculated correctly there will be room for a NULL (0) to
	-- terminate the list of pointers.
	config = allocate_data(ConfigSize)
	NextPtr = config + SIZE_TASKDIALOGCONFIG		-- Store first memory pointer here.
	
	poke4(config,repeat(0,ConfigSize/4))		-- Added /4 after running with SAFE.
	
	-- Store the structure size. The API returns with an error if the size is wrong
	poke4(config,SIZE_TASKDIALOGCONFIG)			-- This length must be correct!
	
	-- Store the handle of the parent window
	hWndInst &= {0,0}					-- Saves checking for atom or short sequence
	poke4( config+hwndParent, hWndInst[1] )
	
	-- Store the instance
	if length(hWndInst) >= 4 then		-- Including the extra {0,0}
		poke4( config+hInstance, hWndInst[2] )
	else
		poke4( config+hInstance, instance() )
	end if

	-- The flags are not stored yet in case a future update wants to chnage them.
	Flags = or_all(Flags)
	
	-- Store the title
	if UTF16( config+pszWindowTitle, Title )=0 then
		FreeConfig(config)
		return { E_INVALIDARG, "Task dialog title is not a valid Unicode string",0,0 }
	end if
	
	-- Store the main instruction
	if UTF16( config+pszMainInstruction, MainText ) = 0 then
		FreeConfig(config)
		return { E_INVALIDARG, "Main Instruction is not a valid Unicode string",0,0 }
	end if
	
	-- Store the content, if any
	-- Is it just a string without a width?
	if UTF16( config+pszContent, Content_W) then
		poke4(config+cxWidth,0)
	else		-- No, so it should be {Content,{Width}}		
		if length(Content_W)=2		-- An atom would have satisfied the first condition
		and UTF16( config+pszContent, Content_W[1])
		and sequence(Content_W[2])
		and integer(Content_W[2][1]) then
			poke4(config+cxWidth,Content_W[2][1])
		else
			FreeConfig(config)
			return { E_INVALIDARG, "Invalid Content_W parameter",0,0 }
		end if
	end if

	-- Store Footer
	if UTF16( config+pszFooter, Footer ) = 0 then
		FreeConfig(config)
		return { E_INVALIDARG, "Footer text is not a valid Unicode string",0,0 }
	end if
	
	-- Store the main icon	
	flg = and_bits(Flags, TDF_USE_HICON_MAIN)
	Icons &= {0,0}
	if sequence(Icons[1]) then
		if flg then
			FreeConfig(config)
			return { E_INVALIDARG, "TDF_USE_HICON_MAIN should not be set with Icon ID",0,0 }
		else		-- Or could set the flag !!!
			poke4( config+pszMainIcon, Icons[1][1] )
		end if
	else
		if flg=0 and Icons[1] != 0 then
			FreeConfig(config)
			return { E_INVALIDARG, "TDF_USE_HICON_MAIN should be set with Icon handle",0,0 }
		else		-- Or could clear the flag !!!
			poke4( config+hMainIcon, Icons[1] )
		end if
	end if

	-- Store the footer icon	
	flg = and_bits(Flags, TDF_USE_HICON_FOOTER)
	if sequence(Icons[2]) then
		if flg then
			FreeConfig(config)
			return { E_INVALIDARG, "TDF_USE_HICON_FOOTER should not be set with Icon ID",0,0 }
		else		-- Or could set the flag !!!
			poke4( config+pszFooterIcon, Icons[2] )
		end if
	else
		if flg=0 and Icons[2] != 0 then
			FreeConfig(config)
			return { E_INVALIDARG, "TDF_USE_HICON_FOOTER should be set with Icon handle",0,0 }
		else		-- Or could clear the flag !!!
			poke4( config+hFooterIcon, Icons[2] )
		end if
	end if
	
	-- Accumulate common button flags and store the default button and custom buttons
	flg = 0
	cBtn = 0		-- Counts custom buttons
	if length(Buttons) then
		Bptr = allocate(length(Buttons)*8+2)			-- Maybe a few more than needed
		poke4(NextPtr,Bptr)
		NextPtr += 4
		poke4(config+pButtons,Bptr)
	else
		poke4(config+pButtons,0)
	end if
	poke4(config+nDefaultButton,0)
	for n = 1 to length(Buttons) do
		if atom(Buttons[n]) then
			flg = or_bits(flg,Buttons[n])
		elsif length(Buttons[n])=1 then
			poke4(config+nDefaultButton,Buttons[n][1])
		elsif length(Buttons[n])=2 then
			cBtn += 1
			poke4(Bptr,Buttons[n][1])
			if UTF16(Bptr+4,Buttons[n][2]) = 0 then
				FreeConfig(config)
				return {E_INVALIDARG,"Buttons argument is invalid",0,0}
			end if
			Bptr += 8
		else
			FreeConfig(config)
			return {E_INVALIDARG,"Buttons argument is invalid",0,0}
		end if
		poke2(Bptr,0)		-- Not required, but it assists diagnostics
	end for
	-- Store the common button flags and the custom button count
	poke4(config+dwCommonButtons,flg)
	poke4(config+cButtons,cBtn)

	-- Store the default radio button and the radio buttons
	cBtn = 0		-- Counts radio buttons
	if length(Radios) then
		Bptr = allocate(length(Radios)*8+2)			-- Maybe a few more than needed
		poke4(NextPtr,Bptr)
		NextPtr += 4
		poke4(config+pRadioButtons,Bptr)
	else
		poke4(config+pRadioButtons,0)
	end if
	poke4(config+nDefaultRadioButton,0)
	for n = 1 to length(Radios) do
		if atom(Radios[n]) then
			poke4(config+nDefaultRadioButton,Radios[n])
		elsif length(Radios[n])=1 then
			poke4(config+nDefaultRadioButton,Radios[n][1])
		elsif length(Radios[n])=2 then
			cBtn += 1
			poke4(Bptr,Radios[n][1])
			if UTF16(Bptr+4,Radios[n][2]) = 0 then
				FreeConfig(config)
				return {E_INVALIDARG,"Radios argument is invalid",0,0}
			end if
			Bptr += 8
		else
			FreeConfig(config)
			return {E_INVALIDARG,"Radios argument is invalid",0,0}
		end if
		poke2(Bptr,0)		-- Not required, but it assists diagnostics
	end for
	-- Store the radio button count
	poke4(config+cRadioButtons,cBtn)

	-- Store Expanded text and controls
	if length(ExpandCC)>3 then
		FreeConfig(config)
		return { E_INVALIDARG, "ExpandCC should be a sequence of no more than three strings",0,0 } 
	end if
	ExpandCC &= {0,0,0}		-- Make sure all fields are available 
	if UTF16(config+pszExpandedInformation,ExpandCC[1])=0 then
		FreeConfig(config)
		return {E_INVALIDARG, "Invalid expanded information in ExpandCC",0,0 }
	end if
	if UTF16(config+pszExpandedControlText,ExpandCC[3])=0 then
		FreeConfig(config)
		return {E_INVALIDARG, "Invalid expanded control text in ExpandCC",0,0 }
	end if
	if UTF16(config+pszCollapsedControlText,ExpandCC[2])=0 then
		FreeConfig(config)
		return {E_INVALIDARG, "Invalid collapsed control text in ExpandCC",0,0 }
	end if

	-- Store the Verification text
	if UTF16(config+pszVerificationText,Verify)=0 then
		FreeConfig(config)
		return {E_INVALIDARG, "Invalid verification text",0,0}
	end if

	-- Store the callback and its data
	CallBack &= {0,0,0}
	if CallBack[1] then
		poke4(config+pfCallback, call_back(CallBack[1]))
	else
		poke4(config+pfCallback, 0)
	end if
	poke4(config+lpCallbackData, CallBack[2])

	-- Store the flags. Stored last to allow future updates to set or clear flags.
	poke4( config+dwFlags, Flags )
	
	if CallBack[3] then
		return config
	end if

	Result = TaskDialogIndirect(config,0)		-- Use ,1) later

	FreeConfig(config)
	return Result
end function

--****
-- === Callback support
--
-- A callback routine is an application-defined function called from a task dialog when
-- various events occur. To register a callback routine, use the ##CallBack## parameter
-- in [[:TaskDialog]] routine. In the full form {##RID##,##CBdata##,##DontDisplay##}~:
-- * ##RID## is the Euphoria routine_id of the callback routine.
-- * ##CBdata## is a pointer to a data area defined by the application, or 0
-- if there is no data.
-- * ##DontDisplay## would be only very rarely used; probably only to implement
-- [[:Multipage dialogs]]. If set, the dialog is not displayed immediately. Instead
-- the address of an extended task dialog configuration structure is returned so
-- that the function [[:TaskDialogIndirect]] can be used to display it later.
-- 
-- A typical header for the callback function could be~:
-- <eucode>
-- function MyCallback(atom hwnd, atom uNotification, atom wParam, atom lParam, dwRefData)
-- </eucode>
-- The task dialog calls this function with the following values~:
-- # ##hwnd## is the handle of the task dialog window
-- # ##uNotification## is one of the notifications in [[:Callback notifications]]
-- # ##wParam## depends upon the notification, as does
-- # ##lParam##. See the appropriate notification for more information.
-- # ##dwRefData## is the ##CBdata## parameter from the task dialog. 
--
-- The callback function should return 0 except as indicated in the notifications
-- [[:TDN_BUTTON_CLICKED]] and [[:TDN_TIMER]]\\
-- For more information refer to the Microsoft documentation
-- [[TaskDialogCallbackProc -> http://msdn.microsoft.com/en-us/library/windows/desktop/bb760542(v=vs.85).aspx]]
--
-- 

--**
-- Used to send a message from a task dialog callback routine to the task dialog.
-- Messages for task dialogs are listed in [[:Task dialog messages]]
--
-- Parameters:
--
-- # ##hWnd## : The handle of the task dialog window = ##hWnd## of the callback routine
-- # ##Msg## : The message to be sent to the task dialog
-- # ##wParam## : This, and ##lParam##, are dependent upon the message ##Msg##.
-- # ##lParam## : This may an atom, restricted to a 32-bit integer, or a sequence, 
-- which is stored as a null-terminated string. A pointer to the string is then used 
-- in the SendMessageW API.
--
-- Returns:
--
-- See the descriptions of the messages [[:TDM_SET_PROGRESS_BAR_STATE]],
-- [[:TDM_SET_PROGRESS_BAR_RANGE]] and [[:TDM_SET_PROGRESS_BAR_POS]].
-- For all other task dialog messages the return value is irrelevant.
--
-- This is a general routine which may also be useful elsewhere.
--

public function SendMessageW(atom hWnd, integer Msg, integer wParam=0, object lParam=0)
atom lPar, rtn
	lPar = UTF16(0,lParam)		-- This returns atoms unchanged.
	rtn = c_func(SendMsg,{hWnd,Msg,wParam,lPar})
	if sequence(lParam) then
		free(lPar)
	end if
	return rtn
end function

--**
-- Releases the memory allocated to a task dialog structure that was created using the
-- TaskDialog function. Usually called from [[:TaskDialogIndirect]] which, in turn, is
-- usually called from [[:TaskDialog]]. Do not call this procedure to release a 
-- structure that was not created by TaskDialog.
--
-- Parameter:
--
-- ##config## : A pointer to an extended task dialog structure
--

export procedure FreeConfig(atom config)
atom ptr, ptrptr
	ptrptr = config + SIZE_TASKDIALOGCONFIG
	while 1 do
		ptr = peek4u(ptrptr)
		if ptr then
			free(ptr)
			ptrptr += 4
		else
			exit
		end if
	end while
	free(config)
end procedure

--**
-- Displays a task dialog as defined by a TASKDIALOGCONFIG structure created by the 
-- [[:TaskDialog]] routine and almost always called directly from it.
--
-- The only circumstances in which a program would call this routine directly is if
-- one or more task dialog configuration structures have been created in advance
-- without displaying them; perhaps to create a multi-page dialog that can be
-- displayed using TDM_NAVIGATE_PAGE messages.
--
-- Parameters:
--
-- # ##config## : A TASKDIALOGCONFIG structure which would normally have been
-- constructed by the TaskDialog function.
-- # ##Clean## : Set TRUE (default) to free the memory used by a config structure that
-- was created by TaskDialog. This should be FALSE (0, the default) if the structure was
-- created by any other means, since the structure created by TaskDialog is extended
-- to include pointers to additional memory to be freed.
--
-- Returns:
--
-- The results from the TaskDialogIndirect API; a sequence of four elements.
--
-- # ##ReturnCode## : One of the following [[:Return Codes]]
-- ** S_OK
-- ** E_OUTOFMEMORY
-- ** E_INVALIDARG
-- ** E_FAIL
-- # ##ButtonID## : The button ID of a selected button, which may be one of the custom
-- buttons or one of the following common buttons: see [[:Return codes]]
-- ** IDCANCEL
-- ** IDNO
-- ** IDOK
-- ** IDRETRY
-- ** IDYES
-- # ##Radio## : The button ID of a selected radio button, or 0 if there are no radio
-- buttons or none have been selected.
-- # ##Verified## : True if the verification checkbox is set, False if it is clear or
-- there is no checkbox
--

export function TaskDialogIndirect(atom config, integer Clean=0)
atom OutPtrs
sequence result
	OutPtrs = allocate(12,1)		-- Could be a permanent location ?
	result =  { c_func(TaskDlgInd,{ config, OutPtrs, OutPtrs+4, OutPtrs+8} ) }
	result &= peek4u({OutPtrs,3})
	if Clean then
		FreeConfig(config)
	end if
	return result 
end function

--	Everything after this is constants.

--****
-- === Task Dialog Flags
--
-- The descriptions below are brief. More detailed descriptions of these constants may
-- be found in the original Microsoft documentation for the
-- [[TASKDIALOGCONFIG -> http://msdn.microsoft.com/en-us/library/windows/desktop/bb787473(v=vs.85).aspx ]]
-- structure.
--

export constant
	--** Enables hyperlink processing by the callback routine. Be wary of unsafe links
	TDF_ENABLE_HYPERLINKS = #0001,
	--** Set this flag if a main icon is specified via a handle
	TDF_USE_HICON_MAIN = #0002,
	--** Set this flag if a footer icon is specified via a handle
	TDF_USE_HICON_FOOTER = #0004,	
	--** Allow dialog cancellation using Alt-F4, Escape and the title bar's close button
	TDF_ALLOW_DIALOG_CANCELLATION = #0008,
	--** Diplays custom buttons as command links with a standard task dialog glyph instead 
	-- of push buttons. Use them when more information is to be associated with the button. 
	-- All characters up to the first new line character will be treated as the link's 
	-- main text. The remainder will be displayed under the main text.
	TDF_USE_COMMAND_LINKS = #0010,
	--** //Not recommended//. Displays custom buttons as command links instead of push 
	-- buttons, but without a glyph. See [[:TDF_USE_COMMAND_LINKS]] for more information.
	TDF_USE_COMMAND_LINKS_NO_ICON = #0020,
	--** Expand the footer area instead of the content (if there is expanded informaton)
	TDF_EXPAND_FOOTER_AREA = #0040,
	--** Expanded information is displayed when the task dialog is initially displayed
	TDF_EXPANDED_BY_DEFAULT = #0080,
	--** The verification checkbox, if present, will be checked when the task dialog 
	-- is initially displayed
	TDF_VERIFICATION_FLAG_CHECKED = #0100,
	--** A progress bar is to be displayed
	TDF_SHOW_PROGRESS_BAR = #0200,
	--** A marquee progress bar is to be displayed
	TDF_SHOW_MARQUEE_PROGRESS_BAR = #0400,
	--** The callback routine is to be called approximately every 200 milliseconds
	TDF_CALLBACK_TIMER = #0800,
	--** The task dialog is centered to the window specified as the parent. If this flag 
	-- is not set, or no parent is specified, the task dialog is centered in the screen
	TDF_POSITION_RELATIVE_TO_WINDOW = #1000,
	--** Indicates that text is to be read right to left
	TDF_RTL_LAYOUT = #2000,
	--** Indicates that no default radio button will be selected
	TDF_NO_DEFAULT_RADIO_BUTTON = #4000,
	--** Allows the task dialog to be minimized
	TDF_CAN_BE_MINIMIZED = #8000,
	--** Deprecated. Use [[:TDF_SIZE_TO_CONTENT]] instead
	TDIF_SIZE_TO_CONTENT = #10000000,
	--** Adjust the width of the task dialog according to the content area, if present, 
	-- unless a width is specified in the TaskDialog argument ##Content_W##
	TDF_SIZE_TO_CONTENT = #1000000
	
TDF_default = TDF_SIZE_TO_CONTENT



--****
-- === Common button flags
--
export constant
	--** The task dialog contains the push button **OK**. If no common or custom buttons 
	-- are specified the dialog will contain an **OK** button by default 
	TDCBF_OK_BUTTON = #01,
	--** The task dialog contains the push button **Yes**
	TDCBF_YES_BUTTON = #02,
	--** The task dialog contains the push button **No**
	TDCBF_NO_BUTTON = #04,
	--** The task dialog contains the push button **Cancel**. If this button is specified 
	-- the task dialog will respond to typical cancel actions (Alt-F4 and Escape)
	TDCBF_CANCEL_BUTTON = #08,
	--** The task dialog contains the push button **Retry**
	TDCBF_RETRY_BUTTON = #10,
	--** The task dialog contains the push button **Close**
	TDCBF_CLOSE_BUTTON = #20

--****
-- === Predefined icons
--
export constant
	--** An exclamation-point icon appears in the task dialog
	TD_WARNING_ICON = #FFFF, 
	--** A stop-sign icon appears in the task dialog
	TD_ERROR_ICON = #FFFE,
	--** An icon consisting of a lower case i in a circle icon appears in the task dialog
	TD_INFORMATION_ICON = #FFFD,
	--** A shield icon appears in the task dialog
	TD_SHIELD_ICON = #FFFC

--****
-- === Return codes
--
export constant
	--** Successful
	S_OK = #00000000,
	--** Returned by a callback routine to indicate special action required.
	S_FALSE = 1,
	--** Insufficient memory to display the task dialog
	E_OUTOFMEMORY = #800700E,
	--** One or more arguments are invalid
	E_INVALIDARG = #80070057,
	--** Task dialog failed for an unknown reason
	E_FAIL = #80004005,
	--** The **Cancel** button was pressed
	IDCANCEL = 2,
	--** The **NO** button was pressed
	IDNO = 7,
	--** The **OK** button was pressed
	IDOK = 1,
	--** The **Retry** button was pressed
	IDRETRY = 4,
	--** The **Yes** button was pressed
	IDYES = 6,
	--** The **Close** button was pressed
	IDCLOSE = 8

--****
-- === Callback notifications
--
-- The task dialog calls the application's callback routine with these notifications.
-- In notifications where the values of wParam and lParam are not mentioned
-- they are 0.
--
export constant
	--** wParam is the command ID of a button that has been clicked. lParam is 0. 
	-- If the callback routine returns S_FALSE (1) the task dialog will continue 
	-- instead of its normal action, which is to finish and return.
    TDN_BUTTON_CLICKED          = 2,
	--** A second indication that the task dialog has been created following the 
	-- TDN_DIALOG_CONSTRUCTED notification. This notification can usually be ignored. 
	-- The task dialog has still not been displayed but would be displayed on return 
	-- unless a default radio button is to be set.
    TDN_CREATED                 = 0,
	--** Indicates that the task dialog has been destroyed.\\
	-- Do not send any more messages to it!
    TDN_DESTROYED               = 5,
	--** This is the first notification received by the callback routine and indicates 
	-- that the task dialog has been created and messages may be sent to it to modify 
	-- its appearance. The dialog has not yet been displayed.
    TDN_DIALOG_CONSTRUCTED      = 7,
	--** Indicates that the button which expands or collapses the expanded information 
	-- has been clicked. wParam is True (1) if the information is expanded, 
	-- False (0) if it is collapsed. lParam is 0. 
    TDN_EXPANDO_BUTTON_CLICKED  = 10,
	--** Indicates that the F1 key has been pressed while the task dialog has focus.
    TDN_HELP                    = 9,
	--** Indicates that a hyperlink has been selected. lParam is a pointer to a 
	-- unicode string with the URL of the hyperlink. wParam is 0. 
    TDN_HYPERLINK_CLICKED       = 3,
	--** "A navigation has occurred". This indicates that the previous task dialog has 
	-- been replaced by a new one.
    TDN_NAVIGATED               = 1,
	--** Indicates that a radio button has been clicked. wParam is its ID\\
	-- This notification is also received just before the dialog is displayed when 
	-- a default radio button is set.
    TDN_RADIO_BUTTON_CLICKED    = 6,
	--** Indicates that a time interval has elapsed. wParam is the total time in 
	-- milliseconds since the task dialog was created or the timer was reset. 
	-- The timer can be rest by returning S_FALSE from the callback routine.
    TDN_TIMER                   = 4,
	--** Indicates that the state of the verification checkbox has been changed 
	-- wParam is True (1) if it is checked and False (0) otherwise.
    TDN_VERIFICATION_CLICKED    = 8


--****
-- === Task dialog messages
--
-- A callback routine can send these messages to the task dialog using SendMessageW. 
-- The descriptions are incomplete. Consult the Microsoft documentation for each 
-- message for more information.
--
export constant
	--** Replaces the existing task dialog with a new one. wParam must be 0 and \\
	-- lParam must be a pointer to a new TASKDIALOGCONFIG structure. 
    TDM_NAVIGATE_PAGE                       = WM_USER + 101,
	--** Simulates a click of the button identified by wParam. lParam must be 0.\\ 
	-- The callback routine receives a TDN_BUTTON_CLICKED notification for it.
    TDM_CLICK_BUTTON                        = WM_USER + 102,
	--** Show the progress bar in marquee mode if wParam is true. False turns off 
	-- marquee mode. lParam must be 0.
    TDM_SET_MARQUEE_PROGRESS_BAR            = WM_USER + 103,
	--** lParam must be 0. \\wParam may be 1 to set the normal state, 2 to set an error 
	-- state or 3 to set a paused state.\\Returns non-zero if successful, 0 if it fails.
    TDM_SET_PROGRESS_BAR_STATE              = WM_USER + 104,
	--** wParam must be 0.\\
	-- The lower 16 bits of lParam specifies the minimum. The default value is 0.\\
	-- The higher 16 bits of lParam specify the maximum. The default value is 100.\\
	-- Returns the previous minimum and maximum in the same format.
    TDM_SET_PROGRESS_BAR_RANGE              = WM_USER + 105,
	--** lParam must be 0. \\wParam is an integer specifying the new position.\\
	-- Returns with the previous position.
    TDM_SET_PROGRESS_BAR_POS                = WM_USER + 106,
	--** wParam is True to turn the marquee display on, False to turn it off. \\
	-- lParam specifies the time in milliseconds between animation updates. 
	-- If 0, it is set to 30 milliseconds.
    TDM_SET_PROGRESS_BAR_MARQUEE            = WM_USER + 107,
	--** Updates a text element in a dialog. 
	-- The size of the task dialog may change to accommodate the new text.\\
	-- lParam is a unicode string (or a pointer to a pre-allocated unicode string).\\ 
	-- wParam is one of the values [[:TDE_CONTENT]], [[:TDE_EXPANDED_INFORMATION]], 
	-- [[:TDE_FOOTER]] or [[:TDE_MAIN_INSTRUCTION]] to indicate which 
	-- text is to be updated.
    TDM_SET_ELEMENT_TEXT                    = WM_USER + 108,
	--** Simulates a click of the radio button identified by wParam. lParam must be 0.
	-- The callback routine receives a TDN_RADIO_BUTTON_CLICKED notification for it.
    TDM_CLICK_RADIO_BUTTON                  = WM_USER + 110,
	--** Enables or disables the button whose ID is in wParam.\\
	-- lParam is 0 to disable the button, non-zero to enable it.
    TDM_ENABLE_BUTTON                       = WM_USER + 111,
	--** Enables or disables the radio button whose ID is in wParam.\\
	-- lParam is 0 to disable the radio button, non-zero to enable it.
    TDM_ENABLE_RADIO_BUTTON                 = WM_USER + 112,
	--** wParam is true (1) to set the verification checkbox, false (0) to clear it.\\
	-- lParam is true (1) to set the focus to the checkbox,
	-- false (0) to leave the focus unaltered.
    TDM_CLICK_VERIFICATION                  = WM_USER + 113,
	--** Updates a text element in a dialog. WARNING: The size of the task dialog 
	-- will not be changed and longer text may be clipped!\\
	-- lParam is a unicode string (or a pointer to a pre-allocated unicode string).\\ 
	-- wParam is one of the values [[:TDE_CONTENT]], [[:TDE_EXPANDED_INFORMATION]], 
	-- [[:TDE_FOOTER]] or [[:TDE_MAIN_INSTRUCTION]] to indicate which 
	-- text is to be updated.
    TDM_UPDATE_ELEMENT_TEXT                 = WM_USER + 114,
	--** Specifies whether a button or command link should have a User Account Control 
	-- (UAC) shield icon; that is whether it requires 'elevation'. \\
	-- wParam is the ID of the button. \\lParam is non-zero to require elevation, 
	-- 0 to specify no elevation.
    TDM_SET_BUTTON_ELEVATION_REQUIRED_STATE = WM_USER + 115,
	--** wParam is 0 to update the main icon, 1 to update the footer icon.\\
	-- lParam identifies the new icon by means of a handle, if TDF_USE_HICON_MAIN is 
	-- set and otherwise one of the predefined icons.
    TDM_UPDATE_ICON                         = WM_USER + 116,
    --** Used in a TDM_SET_ELEMENT_TEXT or TDM_UPDATE_ELEMENT_TEXT message to indicate 
    -- that the Content text is to be changed
    TDE_CONTENT = 0,
    --** Used in a TDM_SET_ELEMENT_TEXT or TDM_UPDATE_ELEMENT_TEXT message to indicate 
    -- that the expanded information is to be changed
    TDE_EXPANDED_INFORMATION = 1,
    --** Used in a TDM_SET_ELEMENT_TEXT or TDM_UPDATE_ELEMENT_TEXT message to indicate 
    -- that the footer text is to be changed
    TDE_FOOTER = 2,
    --** Used in a TDM_SET_ELEMENT_TEXT or TDM_UPDATE_ELEMENT_TEXT message to indicate 
    -- that the main instruction is to be changed
    TDE_MAIN_INSTRUCTION = 3

--****
-- === Multi-page dialogs
--
-- To create a multi-page task dialog, create the task dialog configurations using 
-- [[:TaskDialog]] with the ##CallBack## subparameter ##DontDisplay## set to non-zero.
-- TaskDialog then returns the address of the configuraion structure instead of
-- displaying the task dialog. Create the first page using [[:TaskDialogIndirect]] 
-- and switch to later pages using the [[:TDM_NAVIGATE_PAGE]] message in the callback
-- routine. Free the memory for the configurations using [[:FreeConfig]].


-- Not for the DOCs:

-- The TaskDialogIndirect API and the TASKDIALOGCONFIG structure do not mention
-- Unicode directly. However the more basic routine TaskDialog does specify Unicode
-- and the argument type PCWSTR in the TASKDIALOGCONFIG structure specifies Unicode.

-- UTF-32 characters beyond the UTF-16 range will be translated into the appropriate
-- surrogate pairs by untested code. However, the typeface used by task dialogs,
-- Segoe UI, cannot display them except as the rectangle used to display unknowns. 

-- Strings of less than three characters are unlikely to be required by anyone except
-- someone looking for holes. The purpose of the condition is to avoid confusion
-- between a parameter "me" and a parameter such as {pointer, icon handle}.
-- However, short strings can be used by using {"Hi"} or {"Hi",icon} for example.

 
 
397.0